@@ -8,24 +8,15 @@
  *****************************************************************************/
 #include "keychain_p.h"
 #include "gnomekeyring_p.h"
-
-#include <QSettings>
+#include "libsecret_p.h"
+#include "plaintextstore_p.h"
 
 #include <QScopedPointer>
 
 using namespace QKeychain;
 
-static QString typeKey( const QString& key )
-{
-    return QString::fromLatin1( "%1/type" ).arg( key );
-}
-
-static QString dataKey( const QString& key )
-{
-    return QString::fromLatin1( "%1/data" ).arg( key );
-}
-
 enum KeyringBackend {
+    Backend_LibSecretKeyring,
     Backend_GnomeKeyring,
     Backend_Kwallet4,
     Backend_Kwallet5
@@ -44,7 +35,7 @@ enum DesktopEnvironment {
 // licensed under BSD, see base/nix/xdg_util.cc
 
 static DesktopEnvironment getKdeVersion() {
-    QString value = qgetenv("KDE_SESSION_VERSION");
+    QByteArray value = qgetenv("KDE_SESSION_VERSION");
     if ( value == "5" ) {
         return DesktopEnv_Plasma5;
     } else if (value == "4" ) {
@@ -85,26 +76,82 @@ static DesktopEnvironment detectDesktopEnvironment() {
     return DesktopEnv_Other;
 }
 
+static bool isKwallet5Available()
+{
+    if (!QDBusConnection::sessionBus().isConnected())
+        return false;
+
+    org::kde::KWallet iface(
+        QLatin1String("org.kde.kwalletd5"),
+        QLatin1String("/modules/kwalletd5"),
+        QDBusConnection::sessionBus());
+
+    // At this point iface.isValid() can return false even though the
+    // interface is activatable by making a call. Hence we check whether
+    // a wallet can be opened.
+
+    iface.setTimeout(500);
+    QDBusMessage reply = iface.call(QStringLiteral("networkWallet"));
+    return reply.type() == QDBusMessage::ReplyMessage;
+}
+
 static KeyringBackend detectKeyringBackend()
 {
+    /* The secret service dbus api, accessible through libsecret, is supposed
+     * to unify password services.
+     *
+     * Unfortunately at the time of Kubuntu 18.04 the secret service backend
+     * in KDE is gnome-keyring-daemon - using it has several complications:
+     * - the default collection isn't opened on session start, so users need
+     *   to manually unlock it when the first application uses it
+     * - it's separate from the kwallet5 keyring, so switching to it means the
+     *   existing keyring data can't be accessed anymore
+     *
+     * Thus we still prefer kwallet backends on KDE even if libsecret is
+     * available.
+     */
+
     switch (detectDesktopEnvironment()) {
     case DesktopEnv_Kde4:
         return Backend_Kwallet4;
-        break;
+
     case DesktopEnv_Plasma5:
+        if (isKwallet5Available()) {
+            return Backend_Kwallet5;
+        }
+        if (LibSecretKeyring::isAvailable()) {
+            return Backend_LibSecretKeyring;
+        }
+        if (GnomeKeyring::isAvailable()) {
+            return Backend_GnomeKeyring;
+        }
+        // During startup the keychain backend might just not have started yet
         return Backend_Kwallet5;
-        break;
-        // fall through
+
     case DesktopEnv_Gnome:
     case DesktopEnv_Unity:
     case DesktopEnv_Xfce:
     case DesktopEnv_Other:
     default:
-        if ( GnomeKeyring::isAvailable() ) {
-            return Backend_GnomeKeyring;
-        } else {
-            return Backend_Kwallet4;
+        if (LibSecretKeyring::isAvailable()) {
+            return Backend_LibSecretKeyring;
         }
+        if (GnomeKeyring::isAvailable()) {
+            return Backend_GnomeKeyring;
+        }
+        if (isKwallet5Available()) {
+            return Backend_Kwallet5;
+        }
+        // During startup the keychain backend might just not have started yet
+        //
+        // This doesn't need to be libsecret because LibSecretKeyring::isAvailable()
+        // only fails if the libsecret shared library couldn't be loaded. In contrast
+        // to that GnomeKeyring::isAvailable() can return false if the shared library
+        // *was* loaded but its libgnome_keyring::is_available() returned false.
+        //
+        // In the future there should be a difference between "API available" and
+        // "keychain available".
+        return Backend_GnomeKeyring;
     }
 
 }
@@ -133,6 +180,11 @@ static void kwalletReadPasswordScheduledStartImpl(const char * service, const ch
 
 void ReadPasswordJobPrivate::scheduledStart() {
     switch ( getKeyringBackend() ) {
+    case Backend_LibSecretKeyring: {
+        if ( !LibSecretKeyring::findPassword(key, q->service(), this) ) {
+            q->emitFinishedWithError( OtherError, tr("Unknown error") );
+        }
+    } break;
     case Backend_GnomeKeyring:
         this->mode = JobPrivate::Text;
         if ( !GnomeKeyring::find_network_password( key.toUtf8().constData(),
@@ -217,15 +269,16 @@ void JobPrivate::gnomeKeyring_readCb( int result, const char* string, JobPrivate
 
 void ReadPasswordJobPrivate::fallbackOnError(const QDBusError& err )
 {
-    QScopedPointer<QSettings> local( !q->settings() ? new QSettings( q->service() ) : 0 );
-    QSettings* actual = q->settings() ? q->settings() : local.data();
+    PlainTextStore plainTextStore( q->service(), q->settings() );
 
-    if ( q->insecureFallback() && actual->contains( dataKey( key ) ) ) {
+    if ( q->insecureFallback() && plainTextStore.contains( key ) ) {
+        mode = plainTextStore.readMode( key );
+        data = plainTextStore.readData( key );
 
-        mode = JobPrivate::stringToMode( actual->value( typeKey( key ) ).toString() );
-        data = actual->value( dataKey( key ) ).toByteArray();
-
-        q->emitFinished();
+        if ( plainTextStore.error() != NoError )
+            q->emitFinishedWithError( plainTextStore.error(), plainTextStore.errorString() );
+        else
+            q->emitFinished();
     } else {
         if ( err.type() == QDBusError::ServiceUnknown ) //KWalletd not running
             q->emitFinishedWithError( NoBackendAvailable, tr("No keychain service available") );
@@ -238,21 +291,20 @@ void ReadPasswordJobPrivate::kwalletOpenFinished( QDBusPendingCallWatcher* watch
     watcher->deleteLater();
     const QDBusPendingReply<int> reply = *watcher;
 
-    QScopedPointer<QSettings> local( !q->settings() ? new QSettings( q->service() ) : 0 );
-    QSettings* actual = q->settings() ? q->settings() : local.data();
-
     if ( reply.isError() ) {
         fallbackOnError( reply.error() );
         return;
     }
 
-    if ( actual->contains( dataKey( key ) ) ) {
+    PlainTextStore plainTextStore( q->service(), q->settings() );
+
+    if ( plainTextStore.contains( key ) ) {
         // We previously stored data in the insecure QSettings, but now have KWallet available.
         // Do the migration
 
-        data = actual->value( dataKey( key ) ).toByteArray();
-        const WritePasswordJobPrivate::Mode mode = WritePasswordJobPrivate::stringToMode( actual->value( typeKey( key ) ).toString() );
-        actual->remove( key );
+        data = plainTextStore.readData( key );
+        const WritePasswordJobPrivate::Mode mode = plainTextStore.readMode( key );
+        plainTextStore.remove( key );
 
         q->emitFinished();
 
@@ -365,17 +417,23 @@ static void kwalletWritePasswordScheduledStart( const char * service, const char
 
 void WritePasswordJobPrivate::scheduledStart() {
     switch ( getKeyringBackend() ) {
+    case Backend_LibSecretKeyring: {
+        if ( !LibSecretKeyring::writePassword(service, key, service, mode,
+                                              data, this) ) {
+            q->emitFinishedWithError( OtherError, tr("Unknown error") );
+        }
+    } break;
     case Backend_GnomeKeyring: {
         QString type;
         QByteArray password;
 
         switch(mode) {
         case JobPrivate::Text:
-            type = "plaintext";
+            type = QLatin1String("plaintext");
             password = data;
             break;
         default:
-            type = "base64";
+            type = QLatin1String("base64");
             password = data.toBase64();
             break;
         }
@@ -404,19 +462,18 @@ void WritePasswordJobPrivate::scheduledStart() {
 
 void WritePasswordJobPrivate::fallbackOnError(const QDBusError &err)
 {
-    QScopedPointer<QSettings> local( !q->settings() ? new QSettings( q->service() ) : 0 );
-    QSettings* actual = q->settings() ? q->settings() : local.data();
-
     if ( !q->insecureFallback() ) {
         q->emitFinishedWithError( OtherError, tr("Could not open wallet: %1; %2").arg( QDBusError::errorString( err.type() ), err.message() ) );
         return;
     }
 
-    actual->setValue( QString::fromLatin1( "%1/type" ).arg( key ), mode );
-    actual->setValue( QString::fromLatin1( "%1/data" ).arg( key ), data );
-    actual->sync();
+    PlainTextStore plainTextStore( q->service(), q->settings() );
+    plainTextStore.write( key, data, mode );
 
-    q->emitFinished();
+    if ( plainTextStore.error() != NoError )
+        q->emitFinishedWithError( plainTextStore.error(), plainTextStore.errorString() );
+    else
+        q->emitFinished();
 }
 
 void JobPrivate::gnomeKeyring_writeCb(int result, JobPrivate* self )
@@ -433,19 +490,15 @@ void JobPrivate::kwalletOpenFinished( QDBusPendingCallWatcher* watcher ) {
     watcher->deleteLater();
     QDBusPendingReply<int> reply = *watcher;
 
-    QScopedPointer<QSettings> local( !q->settings() ? new QSettings( q->service() ) : 0 );
-    QSettings* actual = q->settings() ? q->settings() : local.data();
-
     if ( reply.isError() ) {
         fallbackOnError( reply.error() );
         return;
     }
 
-    if ( actual->contains( key ) )
-    {
+    PlainTextStore plainTextStore( q->service(), q->settings() );
+    if ( plainTextStore.contains( key ) ) {
         // If we had previously written to QSettings, but we now have a kwallet available, migrate and delete old insecure data
-        actual->remove( key );
-        actual->sync();
+        plainTextStore.remove( key );
     }
 
     const int handle = reply.value();
@@ -488,6 +541,11 @@ void JobPrivate::kwalletFinished( QDBusPendingCallWatcher* watcher ) {
 
 void DeletePasswordJobPrivate::scheduledStart() {
     switch ( getKeyringBackend() ) {
+    case Backend_LibSecretKeyring: {
+        if ( !LibSecretKeyring::deletePassword(key, q->service(), this) ) {
+            q->emitFinishedWithError( OtherError, tr("Unknown error") );
+        }
+    } break;
     case Backend_GnomeKeyring: {
         if ( !GnomeKeyring::delete_network_password(
                  key.toUtf8().constData(), q->service().toUtf8().constData(),
